@@ -30,6 +30,12 @@
 #include <exiv2/image.hpp>
 #include <exiv2/error.hpp>
 
+#include <cmath>
+
+#ifdef _MSC_VER
+#define HUGE_VALF HUGE_VAL
+#endif
+
 using namespace Strigi;
 using namespace std;
 
@@ -106,7 +112,7 @@ private:
     const RegisteredField* typeField;
 
 };
-
+/*
 const string commentFieldName("http://freedesktop.org/standards/xesam/1.0/core#contentComment");
 const string manufacturerFieldName("http://freedesktop.org/standards/xesam/1.0/core#cameraManufacturer");
 const string modelFieldName("http://freedesktop.org/standards/xesam/1.0/core#cameraModel");
@@ -131,7 +137,7 @@ const string jpegQualityFieldName("http://freedesktop.org/standards/xesam/1.0/co
 const string userCommentFieldName("http://freedesktop.org/standards/xesam/1.0/core#userComment");
 const string jpegProcessFieldName("http://freedesktop.org/standards/xesam/1.0/core#compressionAlgorithm");
 const string thumbnailFieldName("content.thumbnail");
-
+*/
 /*
  Register the field names so that the StreamIndexer knows which analyzer
  provides what information.
@@ -170,6 +176,8 @@ JpegEndAnalyzerFactory::registerFields(FieldRegister& r) {
         "http://freedesktop.org/standards/xesam/1.0/core#whiteBalance");
     exifFields["Exif.Photo.MeteringMode"] = r.registerField(
         "http://freedesktop.org/standards/xesam/1.0/core#meteringMode");
+    exifFields["Exif.Photo.ISOSpeedRatings"] = r.registerField(
+        "http://freedesktop.org/standards/xesam/1.0/core#isoEquivalent");
     map<string, const RegisteredField*>::const_iterator i = exifFields.begin();
     for (; i != exifFields.end(); ++i) {
         addField(i->second);
@@ -192,7 +200,7 @@ JpegEndAnalyzerFactory::registerFields(FieldRegister& r) {
     jpegProcessField = r.registerField(
         "http://freedesktop.org/standards/xesam/1.0/core#compressionAlgorithm");
     thumbnailField = r.registerField(
-        "content.thumbnail");
+        "http://strigi.sf.net/ontologies/homeless#contentThumbnail");
     typeField = r.typeField;
 
     addField(colorModeField);
@@ -213,6 +221,33 @@ JpegEndAnalyzer::checkHeader(const char* header, int32_t headersize) const {
         = {0xFF, 0xD8, 0xFF};
     return headersize >= 3 &&  memcmp(header, jpgmagic, 3) == 0;
 }
+namespace {
+
+bool
+fnumberToApertureValue(string& fnumber) {
+    // Convert the F number to the aperture value
+    // return true if the value in fnumber is changed to ApertureValue
+    char* end;
+    double c = strtod(fnumber.c_str(), &end);
+    if (c <= 0 || c == HUGE_VALF || *end == '\0') return false;
+    double d = strtod(end + 1, &end);
+    if (d <= 0 || d == HUGE_VALF) return false;
+    c /= d;
+#ifdef _MSC_VER
+    double apex = log10(2.)/log10(c * c);
+    apex *= 65536;
+    apex  = (apex < 0 ? ceil(apex - 0.5) : floor(apex + 0.5));
+#else
+    double apex = round(log2(c * c) * 65536);
+#endif
+    std::ostringstream aperture;
+    aperture << apex << "/65536";
+    fnumber.assign(aperture.str());
+    return true;
+}
+
+}
+
 signed char
 JpegEndAnalyzer::analyze(AnalysisResult& ar, ::InputStream* in) {
     // parse the jpeg file now
@@ -277,6 +312,15 @@ JpegEndAnalyzer::analyze(AnalysisResult& ar, ::InputStream* in) {
     }
 
     for (Exiv2::ExifData::const_iterator i = exif.begin(); i != exif.end();++i){
+        if (i->key() == "Exif.Photo.FNumber") {
+            string aperture(i->toString());
+            if (fnumberToApertureValue(aperture)) {
+                ar.addValue(
+                   factory->exifFields.find("Exif.Photo.ApertureValue")->second,
+                   aperture);
+            }
+            continue;
+        }
         map<string,const RegisteredField*>::const_iterator f
             = factory->exifFields.find(i->key());
         if (f != factory->exifFields.end() && f->second) {
