@@ -1,6 +1,6 @@
 /* This file is part of Strigi Desktop Search
  *
- * Copyright (C) 2006 Jos van den Oever <jos@vandenoever.info>
+ * Copyright (C) 2006,2009 Jos van den Oever <jos@vandenoever.info>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -18,42 +18,97 @@
  * Boston, MA 02110-1301, USA.
  */
 #include "archivereader.h"
-#include <strigi/strigiconfig.h>
-#include "streambase.h"
-#include "substreamprovider.h"
 #include "sdfinputstream.h"
 #include "tarinputstream.h"
-#include "gzipinputstream.h"
-#include "bz2inputstream.h"
 #include "mailinputstream.h"
 #include "rpminputstream.h"
 #include "arinputstream.h"
 #include "zipinputstream.h"
-#include "lzmainputstream.h"
-#include <vector>
-#include <list>
+#include "archiveentrycache.h"
+#include "listinginprogress.h"
 #include <iostream>
-#include <cstring>
+#include <set>
 
 using namespace std;
 using namespace Strigi;
 
 class ArchiveReader::DirLister::Private {
-public:
+private:
     int pos;
-    const vector<EntryInfo> entries;
-    Private(const vector<EntryInfo>& v) :entries(v) {
-        pos = 0;
+    vector<EntryInfo> entries;
+    ListingInProgress* listing;
+    const ArchiveEntryCache::SubEntry* entry;
+    set<std::string> done;
+public:
+    const std::string url;
+    explicit Private(const vector<EntryInfo>& v, int p = 0)
+        :pos(p), entries(v), listing(NULL), entry(NULL) {
+    }
+    explicit Private(ListingInProgress* l, const std::string& u)
+            :pos(0), listing(l), entry(NULL), url(u) {
+        listing->ref();
+    }
+    ~Private() {
+        if (listing && listing->unref()) delete listing;
+    }
+    void operator=(const Private& a) {
+        pos = a.pos;
+        entries = a.entries;
+        if (listing && listing->unref()) delete listing;
+        listing = a.listing;
+        if (listing) listing->ref();
+        entry = a.entry;
+        done = a.done;
     }
     bool
     nextEntry(EntryInfo& e) {
-        if (pos >= 0 && pos < (int)entries.size()) {
+        if (pos < 0 || (listing == NULL && pos >= (int)entries.size())) {
+            // there are no (more) valid entries
+            pos = -1;
+        } else if (listing == NULL) {
+            // get the next entry from the static list
             e = entries[pos++];
+        } else if (entry == NULL) {
+            // acquire a dynamic list and get the first entry
+            entry = listing->nextEntry(url);
+            if (entry && entry->entries.size()) {
+                e = nextEntry();
+                pos = 1;
+            } else {
+                pos = -1;
+            }
+        } else if (pos < (int)entry->entries.size()
+                || listing->nextEntry(entry)) {
+            // e is now the next entry from the dynamic list
+            e = nextEntry();
         } else {
+            // there are no more entries
             pos = -1;
         }
         return pos != -1;
     }
+private:
+    /**
+     * Get the first entry from entry->entries() that is not in the entries
+     * vector.
+     **/
+    EntryInfo
+    nextEntry() {
+        // we can make this more efficient by sorting a set with the entries
+        // that have already been returned. A set is faster than a vector for
+        // lookups by value.
+        ArchiveEntryCache::SubEntryMap::const_iterator i
+            = entry->entries.begin();
+        do {
+            if (done.find(i->first) == done.end()) {
+                done.insert(i->first);
+                pos++;
+                return i->second->entry;
+            }
+        } while (++i != entry->entries.end());
+        // this should never happen!
+        return i->second->entry;
+    }
 };
 
 ArchiveReader::DirLister::DirLister(Private* d) :p(d) {
@@ -72,176 +127,36 @@ bool
 ArchiveReader::DirLister::nextEntry(EntryInfo& e) {
     return p->nextEntry(e);
 }
-
-/**
- * @brief A cache for entries that have already been found.
- * @internal
- *
- * The cache is stored in a tree, sorted on the path in the URL
- * (with @c / as the delimiter).
- */
-class ArchiveEntryCache {
-public:
-    class SubEntry;
-    typedef std::map<std::string, SubEntry *> SubEntryMap;
-
-    /**
-     * @brief Represents an entry in the cache.
-     */
-    class SubEntry {
-    public:
-        /** The cached information about this entry. */
-        EntryInfo entry;
-        /** A list of subentries of this entry. */
-        //can't define staticly constructed object while object is being defined
-        SubEntryMap entries;
-        /** Constructor */
-        SubEntry() { }
-        /** Destructor */
-        virtual ~SubEntry();
-    };
-    /**
-     * @brief Represents a top-level entry in the cache
-     */
-    class RootSubEntry : public SubEntry {
-    public:
-        /** Constructor */
-        RootSubEntry() :SubEntry() {}
-        /** Whether this entry and its subentries have been indexed yet */
-        bool indexed;
-    };
-
-    ~ArchiveEntryCache() {
-       map<string, RootSubEntry*>::const_iterator i;
-       for (i = cache.begin(); i != cache.end(); ++i) {
-           delete i->second;
-       }
-    }
-
-    /**
-     * @brief The cache itself.
-     *
-     * Contains all the root entries in the cache.
-     */
-    std::map<std::string, RootSubEntry*> cache;
-
-    /**
-     * @brief Find an entry in the cache by its URL
-     *
-     * @param url the url of the entry
-     * @return a pointer to the entry, if it is in the cache,
-     * 0 if the entry is not in the cache
-     */
-    const SubEntry* findEntry(const string& url) const;
-    /**
-     * @brief Finds the root subentry in the cache of the given url,
-     * if it is in the cache
-     *
-     * This finds the RootSubEntry (tree root) that the SubEntry for the
-     * given url either is or should be stored under.  If there is no
-     * suitable RootSubEntry in the cache, cache.end() is returned.
-     *
-     * @param url the url of the object to find the root entry for
-     * @return an iterator to the root entry
-     */
-    map<string, RootSubEntry*>::const_iterator findRootEntry(
-        const string& url) const;
-};
-
-ArchiveEntryCache::SubEntry::~SubEntry() {
-    SubEntryMap::iterator i;
-    for (i=entries.begin(); i!=entries.end(); ++i) {
-        delete i->second;
-    }
+const ArchiveReader::DirLister&
+ArchiveReader::DirLister::operator=(const DirLister& a) {
+    *p = *a.p;
+    return a;
 }
-map<string, ArchiveEntryCache::RootSubEntry*>::const_iterator
-ArchiveEntryCache::findRootEntry(const string& url) const {
-    string n(url);
-    size_t p = n.size();
-    do {
-        map<string, RootSubEntry*>::const_iterator i = cache.find(n);
-        if (i != cache.end()) {
-            // the root entry is in the cache - we're done
-            return i;
-        }
-        // remove the last element in the path, and look for that
-        p = n.rfind('/');
-        if (p != string::npos) {
-            n.resize(p);
-        }
-    } while (p != string::npos);
-    // couldn't find it
-    return cache.end();
-}
-/**
- * Find the entry corresponding to @p url.
- **/
-const ArchiveEntryCache::SubEntry*
-ArchiveEntryCache::findEntry(const string& url) const {
-    // find the root entry that should contain the wanted entry
-    map<string, RootSubEntry*>::const_iterator ei = findRootEntry(url);
-    if (ei == cache.end()) return 0; // no root could be find
-    if (ei->first == url) {
-        // the requested entry is a root entry: we are done
-        return ei->second;
-    }
-
-    // use the path components as keys to find the entry
-    const SubEntry* e = ei->second;
-    size_t p = ei->first.length();
-    string name;
-    do {
-        size_t np = url.find('/', p+1);
-        if (np == string::npos) {
-            name.assign(url, p+1, url.size());
-        } else {
-            name.assign(url, p+1, np-p-1);
-        }
-        SubEntryMap::const_iterator i = e->entries.find(name);
-        if (i == e->entries.end()) {
-            e = 0;
-        } else {
-            e = i->second;
-            p = np;
-        }
-        if (p == url.length()) {
-            return e;
-        }
-    } while(e && p != string::npos);
-
-    return e;
-}
-
 
 /** Private data members and functions for ArchiveReader */
 class ArchiveReader::ArchiveReaderPrivate {
 public:
-    class StreamPtr {
-    public:
-        InputStream* stream;
-        SubStreamProvider* provider;
-        StreamPtr() :stream(0), provider(0) {}
-        StreamPtr(InputStream* s) :stream(s), provider(0) {}
-        StreamPtr(SubStreamProvider* p) :stream(0), provider(p) {}
-        void free();
-    };
-    typedef std::map<InputStream*,
-        std::list<StreamPtr> > openstreamsType;
-    openstreamsType openstreams;
+    /** List of streams that have been opened for openStream() calls. **/
+    typedef std::map<InputStream*, std::list<StreamPtr> > OpenstreamsType;
+    OpenstreamsType openstreams;
+    /** List of objects that can open streams. **/
     std::list<StreamOpener*> openers;
+    /** Cache of EntryInfo object to files that have been read. **/
     ArchiveEntryCache cache;
+    /** Map with pointers to a headerCheck function as key and a factory
+     *  function as value. */
     std::map<bool (*)(const char*, int32_t),
         SubStreamProvider* (*)(InputStream*)> subs;
+    /**
+     * A list of dir listings in progress. Each ListingInProgress has a physical
+     * file in common. The url of the physical file is the key.
+     **/
+    std::map<std::string, ListingInProgress*> listingsInProgress;
 
     std::vector<size_t> cullName(const std::string& url,
         InputStream*& stream) const;
-    SubStreamProvider* positionedProvider(const std::string& url,
-        bool onlyfiles);
+    SubStreamProvider* positionedProvider(const std::string& url);
     InputStream* open(const std::string& url) const;
-    SubStreamProvider* subStreamProvider(
-        InputStream*, std::list<StreamPtr>& streams);
-    static void free(std::list<StreamPtr>& l);
-    int fillEntry(ArchiveEntryCache::SubEntry&e, InputStream*s);
     /**
      * @brief Query known StreamOpeners about an URL.
      *
@@ -254,49 +169,8 @@ public:
     int localStat(const std::string& url, EntryInfo& e);
     ArchiveReaderPrivate();
     ~ArchiveReaderPrivate();
+    ListingInProgress* findListingInProgress(const string& url) const;
 };
-void
-ArchiveReader::ArchiveReaderPrivate::StreamPtr::free() {
-    if (stream) delete stream;
-    if (provider) delete provider;
-}
-/**
- * Add the entry @p se to the cache.
- * The filename of @p se contains the complete path.
- * This is split on '/' into its parts. The parts are used as keys to find
- * the parents of the new entry.
- * Any entry that is not yet in the cache is created as a directory.
- **/
-void
-addEntry(ArchiveEntryCache::SubEntry* parent, ArchiveEntryCache::SubEntry* se) {
-    // split path into components, the components are placed in the array
-    // 'names'. The filename is kept in the filename member of se
-    vector<string> names;
-    string& name = se->entry.filename;
-    string::size_type p = name.find('/');
-    while (p != string::npos) {
-        names.push_back(name.substr(0, p));
-        name.erase(0, p + 1);
-        p = name.find('/');
-    }
-
-    // find the right entry
-    ArchiveEntryCache::SubEntry* child;
-    for (size_t i=0; i<names.size(); ++i) {
-        child = parent->entries[names[i]];
-        if (child == NULL) {
-            // create a new directory entry if it is not yet in the cache
-            child = new ArchiveEntryCache::SubEntry();
-            child->entry.filename = names[i];
-            child->entry.type = EntryInfo::Dir;
-            child->entry.size = 0;
-            parent->entries[names[i]] = child;
-        }
-        parent = child;
-    }
-    // this is what we came for: add the entry to the parent
-    parent->entries[name] = se;
-}
 ArchiveReader::ArchiveReaderPrivate::ArchiveReaderPrivate() {
     typedef std::pair<bool (*)(const char*, int32_t),
         SubStreamProvider* (*)(InputStream*)> SubsPair;
@@ -317,11 +191,17 @@ ArchiveReader::ArchiveReaderPrivate::ArchiveReaderPrivate() {
 ArchiveReader::ArchiveReaderPrivate::~ArchiveReaderPrivate() {
     if (openstreams.size() > 0) {
         cerr << openstreams.size() << " streams were not closed." << endl;
-        openstreamsType::iterator i;
+        OpenstreamsType::iterator i;
         for (i = openstreams.begin(); i != openstreams.end(); ++i) {
             free(i->second);
         }
     }
+    map<string, ListingInProgress*>::const_iterator end
+        = listingsInProgress.end();
+    for (map<string, ListingInProgress*>::const_iterator i
+            = listingsInProgress.begin(); i != end; ++i) {
+        if (i->second->unref()) delete i->second;
+    }
 }
 /**
  * Try to open stream corresponding to a url.
@@ -342,9 +222,13 @@ ArchiveReader::ArchiveReaderPrivate::cullName(const string& url,
     }
     return partpos;
 }
+/**
+ * Obtain a substream provider for which the current stream (currentEntry())
+ * points to the stream specified by @p url.
+ * On failure, 0 is returned.
+ **/
 SubStreamProvider*
-ArchiveReader::ArchiveReaderPrivate::positionedProvider(const string& url,
-        bool onlyfiles) {
+ArchiveReader::ArchiveReaderPrivate::positionedProvider(const string& url) {
     InputStream* stream = 0;
 
     // cull the url until a stream can be opened
@@ -353,20 +237,24 @@ ArchiveReader::ArchiveReaderPrivate::positionedProvider(const string& url,
         return 0;
     }
 
-    // open the substreams until we have opened the complete path
+    // open the substreams until the complete path has been opened
     SubStreamProvider* provider;
     InputStream* substream = stream;
     vector<size_t>::reverse_iterator i;
     list<StreamPtr> streams;
     streams.push_back(stream);
     for (i = partpos.rbegin(); i != partpos.rend(); ++i) {
-        const char* sn = url.c_str() + *i;
-        size_t len = url.length();
-        provider = subStreamProvider(substream, streams);
+        // try to open the stream as a SubStreamProvider
+        provider = subStreamProvider(subs, substream, streams);
         if (provider == 0) {
+            free(streams);
             return 0;
         }
+        // let sn point to the trailing part of the url
+        const char* sn = url.c_str() + *i;
+        size_t len = url.length();
         bool nextstream = false;
+        // try to open the first substream of the current SubStreamProvider
         substream = provider->currentEntry();
         do {
             const EntryInfo& e = provider->entryInfo();
@@ -382,9 +270,11 @@ ArchiveReader::ArchiveReaderPrivate::positionedProvider(const string& url,
                     ++i;
                 } while (i != partpos.rend() && *i < end);
                 if (i == partpos.rend()) {
+                    // success!
                     openstreams[substream] = streams;
                     return provider;
                 }
+                // no match: rewind
                 --i;
             } else {
                 substream = provider->nextEntry();
@@ -398,13 +288,6 @@ ArchiveReader::ArchiveReaderPrivate::positionedProvider(const string& url,
     }
     return 0;
 }
-void
-ArchiveReader::ArchiveReaderPrivate::free(list<StreamPtr>& l) {
-    list<StreamPtr>::iterator i;
-    for (i=l.begin(); i!=l.end(); ++i) {
-        i->free();
-    }
-}
 /**
  * Try with each of the streamopeners to open a stream.
  **/
@@ -417,114 +300,6 @@ ArchiveReader::ArchiveReaderPrivate::open(const string& url) const {
     }
     return stream;
 }
-SubStreamProvider*
-ArchiveReader::ArchiveReaderPrivate::subStreamProvider(
-        InputStream* input, list<ArchiveReaderPrivate::StreamPtr>& streams) {
-    if (input == 0) return 0;
-    InputStream* s = input;
-
-    bool foundCompressedStream;
-    do {
-        foundCompressedStream = false;
-        // check if this is a compressed stream
-        const char* c;
-        int32_t n = s->read(c, 16, 0);
-        s->reset(0);
-        if (BZ2InputStream::checkHeader(c, n)) {
-            InputStream* ns = new BZ2InputStream(s);
-            if (ns->status() == Ok) {
-                foundCompressedStream = true;
-                s = ns;
-                streams.push_back(s);
-            } else {
-                delete ns;
-                s->reset(0);
-            }
-        }
-        n = s->read(c, 2, 0);
-        s->reset(0);
-        if (n >= 2 && c[0] == 0x1f && c[1] == (char)0x8b) {
-            InputStream* ns = new GZipInputStream(s);
-            if (ns->status() == Ok) {
-                foundCompressedStream = true;
-                s = ns;
-                streams.push_back(s);
-            } else {
-                delete ns;
-                s->reset(0);
-            }
-        }
-        n = s->read(c, 2, 0);
-        s->reset(0);
-        if (LZMAInputStream::checkHeader(c, n)) {
-            InputStream* ns = new LZMAInputStream(s);
-            if (ns->status() == Ok) {
-                foundCompressedStream = true;
-                s = ns;
-                streams.push_back(s);
-            } else {
-                delete ns;
-                s->reset(0);
-            }
-        }
-    } while (foundCompressedStream);
- 
-    const char* c;
-    int32_t n = s->read(c, 1024, 0);
-    s->reset(0);
-    SubStreamProvider* ss;
-    map<bool (*)(const char*, int32_t),
-        SubStreamProvider* (*)(InputStream*)>::const_iterator i;
-    for (i = subs.begin(); i != subs.end(); ++i) {
-        // check if the header matches for each substreamprovider
-        if (i->first(c, n)) {
-            // create a new SubStreamProvider
-            ss = i->second(s);
-            if (ss->nextEntry()) {
-                streams.push_back(ss);
-                // return the first substream
-                return ss;
-            }
-            // even though the header was good, this stream has no substream
-            delete ss;
-            s->reset(0);
-            n = s->read(c, 1, 0);
-            s->reset(0);
-        }
-    }
-    free(streams);
-    return 0;
-}
-int
-ArchiveReader::ArchiveReaderPrivate::fillEntry(
-        ArchiveEntryCache::SubEntry& e, InputStream* s) {
-    int nentries = 0;
-    list<StreamPtr> streams;
-    SubStreamProvider* p = subStreamProvider(s, streams);
-    if (!p) return 0;
-    do {
-        ArchiveEntryCache::SubEntry* se = new ArchiveEntryCache::SubEntry();
-        se->entry = p->entryInfo();
-        int nsubentries = fillEntry(*se, p->currentEntry());
-        if (se->entry.size < 0) {
-            // read entire stream to determine it's size
-            InputStream *es = p->currentEntry();
-            const char* c;
-            while (es->read(c, 1, 0) > 0) {}
-            se->entry.size = max(es->size(), (int64_t)0);
-        }
-        addEntry(&e, se);
-        if (nsubentries) {
-            nentries += nsubentries;
-        }
-        nentries++;
-    } while (p->nextEntry());
-    free(streams);
-    if (nentries) {
-        e.entry.type = (EntryInfo::Type)(e.entry.type|EntryInfo::Dir);
-    }
-    return nentries;
-}
 int
 ArchiveReader::ArchiveReaderPrivate::localStat(const std::string& url,
         EntryInfo& e) {
@@ -535,9 +310,19 @@ ArchiveReader::ArchiveReaderPrivate::localStat(const std::string& url,
             if (!(e.type & EntryInfo::File)) {
                 return 0;
             }
+
+            // check if a ListingInProgress points to this url
+            map<string, ListingInProgress*>::const_iterator li =
+                listingsInProgress.find(url);
+            if (li != listingsInProgress.end()) {
+                // use the information in this entry
+                e = li->second->root->entry;
+                return 0;
+            }
+ 
             // check if this file is in the cache
             map<string, ArchiveEntryCache::RootSubEntry*>::const_iterator se
-                = cache.findRootEntry(url);
+                = cache.cache.find(url);
             if (se != cache.cache.end()) {
                 if (se->second->entry.mtime == e.mtime) {
                     e.type = se->second->entry.type;
@@ -549,22 +334,23 @@ ArchiveReader::ArchiveReaderPrivate::localStat(const std::string& url,
                 delete rse;
             }
 
-            // we statted, now we'd like to know if this file has subentries
+            // The file exists, but is it an archive?
             InputStream* s = (*i)->openStream(url);
-            list<ArchiveReaderPrivate::StreamPtr> streams;
-            SubStreamProvider* provider = subStreamProvider(s, streams);
+            list<StreamPtr> streams;
+            SubStreamProvider* provider = subStreamProvider(subs, s, streams);
             if (provider) {
                 // this file contains substreams
                 e.type = (EntryInfo::Type)(EntryInfo::Dir|EntryInfo::File);
                 free(streams);
-
+/*
+                // create an empty entry in the cache
                 ArchiveEntryCache::RootSubEntry* rse = cache.cache[url];
                 if (rse == NULL) {
                     rse = new ArchiveEntryCache::RootSubEntry();
                     cache.cache[url] = rse;
                 }
                 rse->indexed = false;
-                rse->entry = e;
+                rse->entry = e;*/
             }
             delete s;
             return 0;
@@ -579,6 +365,7 @@ ArchiveReader::~ArchiveReader() {
 }
 int
 ArchiveReader::stat(const std::string& url, EntryInfo& e) {
+    // try to stat the url as a physical file
     if (p->localStat(url, e) == 0) return 0;
 
     // check the cache (this assumes dirEntries was already called)
@@ -588,11 +375,14 @@ ArchiveReader::stat(const std::string& url, EntryInfo& e) {
         return 0;
     }
     // try reading the entries from the collection to which this file belongs
-    dirEntries(url);
-    subentry = p->cache.findEntry(url);
-    if (subentry) {
-        e = subentry->entry;
-        return 0;
+    size_t pos = url.rfind('/');
+    if (pos == string::npos) return -1;
+    std::string parenturl(url, 0, pos);
+    ArchiveReader::DirLister dirlister(dirEntries(parenturl));
+    while (dirlister.nextEntry(e)) {
+        if (e.filename == url.c_str()+pos+1) {
+            return 0;
+        }
     }
     return -1;
 }
@@ -601,8 +391,8 @@ ArchiveReader::openStream(const string& url) {
     InputStream* stream = p->open(url);
     if (stream) return stream;
 
-    // open the substreams until have opened the complete path
-    SubStreamProvider* provider = p->positionedProvider(url, true);
+    // open the substreams until the complete path has been opened
+    SubStreamProvider* provider = p->positionedProvider(url);
     if (provider) {
         stream = provider->currentEntry();
     }
@@ -614,13 +404,13 @@ ArchiveReader::addStreamOpener(StreamOpener* opener) {
 }
 void
 ArchiveReader::closeStream(InputStream* s) {
-    ArchiveReaderPrivate::openstreamsType::iterator i(
+    ArchiveReaderPrivate::OpenstreamsType::iterator i(
         p->openstreams.find(s));
     if (i == p->openstreams.end()) {
         delete s;
         return;
     }
-    p->free(i->second);
+    free(i->second);
     p->openstreams.erase(i);
 }
 bool
@@ -629,22 +419,30 @@ ArchiveReader::isArchive(const std::string& url) {
     if (p->localStat(url, e) != 0) {
         return false;
     }
-    return e.type & EntryInfo::File && e.type & EntryInfo::Dir;
+    return e.type & (EntryInfo::File | EntryInfo::Dir);
+}
+std::vector<EntryInfo>
+convert(const ArchiveEntryCache::SubEntry* entry) {
+    std::vector<EntryInfo> v;
+    if (entry == NULL) return v;
+    ArchiveEntryCache::SubEntryMap::const_iterator i;
+    for (i = entry->entries.begin(); i != entry->entries.end(); ++i) {
+        v.push_back(i->second->entry);
+    }
+    return v;
 }
 ArchiveReader::DirLister
 ArchiveReader::dirEntries(const std::string& url) {
     // find the entry in the cache
     const ArchiveEntryCache::SubEntry* subentry = p->cache.findEntry(url);
 
-    // if this is a root entry that was not yet indexed, index it now
-    const ArchiveEntryCache::RootSubEntry* root
-        = dynamic_cast<const ArchiveEntryCache::RootSubEntry*>(subentry);
-    if (root && !root->indexed) {
-        // pretend like it's not in the cache
-        subentry = NULL;
+    // look for a ListingInProgress
+    ListingInProgress* lip = NULL;
+    if (subentry == NULL) {
+        lip = p->findListingInProgress(url);
     }
     std::vector<EntryInfo> v;
-    if (subentry == NULL) {
+    if (subentry == NULL && lip == NULL) {
         // this entry is not in the cache, we try to open it
         InputStream* s = 0;
         vector<size_t> l = p->cullName(url, s);
@@ -661,30 +459,35 @@ ArchiveReader::dirEntries(const std::string& url) {
         EntryInfo e;
         // get the properties of the physical file
         p->localStat(name, e);
-        ArchiveEntryCache::RootSubEntry* se = p->cache.cache[name];
-        if (se == NULL) {
-            se = new ArchiveEntryCache::RootSubEntry();
-            p->cache.cache[name] = se;
-        }
-        se->indexed = true;
-        se->entry = e;
-        p->fillEntry(*se, s);
-        delete s;
-        subentry = p->cache.findEntry(url);
+        lip = new ListingInProgress(p->subs, e, name, s);
+        lip->ref();
+        p->listingsInProgress[name] = lip;
     }
 
-    if (subentry) {
-        ArchiveEntryCache::SubEntryMap::const_iterator i;
-        for (i = subentry->entries.begin(); i != subentry->entries.end();
-                ++i) {
-            v.push_back(i->second->entry);
+    if (lip) {
+        if (lip->isDone()) {
+            p->cache.cache[lip->url] = lip->root;
+            lip->root = 0;
+            p->listingsInProgress.erase(lip->url);
+            if (lip->unref()) delete lip;
+        } else {
+            return DirLister(new DirLister::Private(lip, url));
         }
     }
-    DirLister dl(new DirLister::Private(v));
-    return dl;
+
+    if (subentry == NULL) {
+        subentry = p->cache.findEntry(url);
+    }
+    if (subentry) {
+        v = convert(subentry);
+    }
+    return DirLister(new DirLister::Private(v));
 }
 bool
 ArchiveReader::canHandle(const std::string& url) {
+    // remove parts from the back of url until url matches a physical file
+    // return true of the physical file is an archive, i.e. contains
+    // substreams
     size_t pos = url.rfind('/');
     EntryInfo e;
     int r = p->localStat(url, e);
@@ -694,3 +497,24 @@ ArchiveReader::canHandle(const std::string& url) {
     }
     return r == 0 && e.type & EntryInfo::File && e.type & EntryInfo::Dir;
 }
+ListingInProgress*
+ArchiveReader::ArchiveReaderPrivate::findListingInProgress(const string& url)
+        const {
+    string n(url);
+    size_t p = n.size();
+    do {
+        map<string, ListingInProgress*>::const_iterator i
+            = listingsInProgress.find(n);
+        if (i != listingsInProgress.end()) {
+            // the root entry is in the map - we are done
+            return i->second;
+        }
+        // remove the last element in the path, and look for that
+        p = n.rfind('/');
+        if (p != string::npos) {
+            n.resize(p);
+        }
+    } while (p != string::npos);
+    // couldn't find it
+    return 0;
+}
